
from ZODB.POSException import ConflictError

from five import grok
from zope.interface import Interface
from infrae.wsgi.response import AbortPublication


class TestView(grok.View):
    grok.context(Interface)
    grok.name('test.html')

    def render(self):
        return '<html><h1>Test</h1></html>'


class AbortView(grok.View):
    grok.context(Interface)
    grok.name('abort.html')

    def render(self):
        # This should only done by infrae.wsgi code.
        raise AbortPublication(False)


class ConflictuousView(grok.View):
    grok.context(Interface)
    grok.name('conflict.html')

    def render(self):
        raise ConflictError()


class ConflictAndErrorView(grok.View):
    grok.context(Interface)
    grok.name('failed_conflict.html')

    def render(self):
        if self.request.retry_count:
            raise ValueError('Failed')
        raise ConflictError()



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

import unittest

from zope.interface.verify import verifyObject

from infrae.testing import ZCMLLayer
from infrae.wsgi.testing import TestRequest
from infrae.wsgi.interfaces import IRequest, ITraverser
from infrae.wsgi.interfaces import IVirtualHosting, IAuthenticator
import infrae.wsgi


class RequestTestCase(unittest.TestCase):
    """Test the WSGI request Virtual Host support.
    """
    layer = ZCMLLayer(infrae.wsgi)

    def test_test_request(self):
        request = TestRequest()
        self.assertTrue(IRequest.providedBy(request))
        self.assertEqual(
            request.physicalPathToURL('/root'), 'http://localhost/root')
        self.assertEqual(
            request.getURL(), 'http://localhost')

    def test_plugin_traverser(self):
        request = TestRequest()
        retrieved_plugin = request.get_plugin(ITraverser)
        self.assertIs(retrieved_plugin, None)

        plugin = request.query_plugin(request.application, ITraverser)
        self.assertNotEqual(plugin, None)
        self.assertTrue(verifyObject(ITraverser, plugin))

        retrieved_plugin = request.get_plugin(ITraverser)
        self.assertIs(plugin, retrieved_plugin)

    def test_plugin_virtualhosting(self):
        request = TestRequest()
        retrieved_plugin = request.get_plugin(IVirtualHosting)
        self.assertIs(retrieved_plugin, None)

        plugin = request.query_plugin(request.application, IVirtualHosting)
        self.assertNotEqual(plugin, None)
        self.assertTrue(verifyObject(IVirtualHosting, plugin))
        self.assertEqual(plugin.root, None)

        retrieved_plugin = request.get_plugin(IVirtualHosting)
        self.assertIs(plugin, retrieved_plugin)

    def test_plugin_authenticator(self):
        request = TestRequest()
        retrieved_plugin = request.get_plugin(IAuthenticator)
        self.assertIs(retrieved_plugin, None)

        plugin = request.query_plugin(request.application, IAuthenticator)
        self.assertNotEqual(plugin, None)
        self.assertTrue(verifyObject(IAuthenticator, plugin))

        retrieved_plugin = request.get_plugin(IAuthenticator)
        self.assertIs(plugin, retrieved_plugin)


class VirtualHostingTestCase(unittest.TestCase):
    """Test the virtual hosting plugin functionality.
    """
    layer = ZCMLLayer(infrae.wsgi)

    def test_rewrite_url(self):
        request = TestRequest()
        plugin = request.query_plugin(request.application, IVirtualHosting)
        self.assertNotEqual(plugin, None)

        self.assertEqual(
            plugin.rewrite_url(None, 'http://localhost/root/folder/edit'),
            '/root/folder/edit')
        self.assertEqual(
            plugin.rewrite_url('https://admin', 'http://localhost/root/edit'),
            'https://admin/root/edit')


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(RequestTestCase))
    suite.addTest(unittest.makeSuite(VirtualHostingTestCase))
    return suite



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from five import grok
from zope.interface import Interface
from infrae.wsgi.interfaces import IRequest, IAuthenticator

from ZPublisher.BaseRequest import UNSPECIFIED_ROLES, old_validation
from zExceptions import Unauthorized


class ZopeAuthenticator(grok.MultiAdapter):
    """After traversing, authenticate using Zope method.
    """
    grok.adapts(Interface, IRequest)
    grok.provides(IAuthenticator)
    grok.implements(IAuthenticator)

    def __init__(self, context, request):
        self.context = context
        self.request = request

    def validate(self, content):
        auth = self.request._auth
        groups = content.__allow_groups__
        if hasattr(groups, 'validate'):
            if self.request.roles is UNSPECIFIED_ROLES:
                user = groups.validate(self.request, auth)
            else:
                user = groups.validate(self.request, auth, self.request.roles)
        else:
            user = old_validation(groups, self.request, auth, self.request.roles)
        return user

    def __call__(self, validated_hook=None):
        user = None

        # Authenticate user on the published object if available
        if hasattr(self.context, '__allow_groups__'):
            user = self.validate(self.context)

        index = 0
        if user is None:
            # Or on the first possible parent.
            parents = self.request['PARENTS']
            while index < len(parents):
                if hasattr(parents[index], '__allow_groups__'):
                    user = self.validate(parents[index])
                    if user is not None:
                        break
                index += 1

        if user is None:
            # Not authenticated
            if self.request.roles != UNSPECIFIED_ROLES:
                raise Unauthorized('Please authenticate')
        else:
            # Authenticated
            if validated_hook is not None:
                validated_hook(self.request, user)
            self.request['AUTHENTICATED_USER']=user
            self.request['AUTHENTICATION_PATH']='/'.join(
                self.request.steps[:-index])

        return user



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from infrae.wsgi.response import WSGIResponse
from infrae.wsgi.tests.mockers import MockWSGIStartResponse

import unittest


class ResponseTestCase(unittest.TestCase):

    def setUp(self):
        self.start_response = MockWSGIStartResponse()

    def test_simple(self):
        """Test a simple reply.
        """
        response = WSGIResponse({}, self.start_response)
        response.setBody('<p>Hello</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, [])
        self.assertEqual(response.body, '<p>Hello</p>')

    def test_simple_with_content_type(self):
        """Test a simple reply where the content type have been set.
        """
        response = WSGIResponse({}, self.start_response)
        response.setHeader('Content-type', 'text/html;charset=utf-8')
        response.setBody('<p>Hello</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, [])
        self.assertEqual(response.body, '<p>Hello</p>')

    def test_status(self):
        """Test setting/getting a status.
        """
        response = WSGIResponse({}, self.start_response)
        self.assertEqual(response.getStatus(), 200)
        response.setStatus(404)
        self.assertEqual(response.getStatus(), 404)
        response.setBody('<p>Not found</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '404 Not Found')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '16'),
             ('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, [])

    def test_write(self):
        """Test using the write method of Request object.
        """
        response = WSGIResponse({}, self.start_response)
        response.write('Hello')
        response.write('World')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, ['Hello', 'World'])

    def test_headers(self):
        """Try to set some headers.
        """
        response = WSGIResponse({}, self.start_response)
        response.setHeader('cache-control', 'max-age=1200')
        response.addHeader('x-downloaded', 42)
        self.assertEqual(response.getHeader('X-Downloaded'), 42)
        self.assertEqual(response.getHeader('Cache-control'), 'max-age=1200')
        response.setBody('<p>News</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Cache-Control', 'max-age=1200'),
             ('Content-Length', '11'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('X-Downloaded', '42')])
        self.assertEqual(self.start_response.data, [])

    def test_content_type_and_length(self):
        """Here we set a custom content-type and length.
        """
        response = WSGIResponse({}, self.start_response)
        response.setHeader('content-type', 'text/csv')
        self.assertEqual(response.getHeader('Content-Type'), 'text/csv')
        response.setHeader('content-length', 1024)
        response.setBody('blue;yellow')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '1024'),
             ('Content-Type', 'text/csv')])
        self.assertEqual(self.start_response.data, [])

    def test_content_type_and_length_with_write(self):
        """We here set the content type and length before streaming
        using the write method.
        """
        response = WSGIResponse({}, self.start_response)
        response.setHeader('content-type', 'text/csv')
        self.assertEqual(response.getHeader('Content-Type'), 'text/csv')
        response.setHeader('content-length', 1024)
        response.write('blue;yellow')
        response.write('green;black')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '1024'),
             ('Content-Type', 'text/csv')])
        self.assertEqual(
            self.start_response.data, ['blue;yellow', 'green;black'])

    def test_redirect(self):
        """Test redirect to a URL.
        """
        response = WSGIResponse({}, self.start_response)
        response.redirect('http://infrae.com')
        response.setBody('<p>Redirect to Infrae</p>')
        self.assertEqual(response.getStatus(), 302)
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '302 Moved Temporarily')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '25'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Location', 'http://infrae.com')])
        self.assertEqual(self.start_response.data, [])

    def test_redirect_with_status(self):
        """Test redirect to a URL providing a status.
        """
        response = WSGIResponse({}, self.start_response)
        response.redirect('http://infrae.com', status=301)
        self.assertEqual(response.getStatus(), 301)
        response.setBody('<p>Redirect to Infrae</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '301 Moved Permanently')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '25'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Location', 'http://infrae.com')])
        self.assertEqual(self.start_response.data, [])

    def test_redirect_no_content(self):
        """Test redirect without setting any content.
        """
        response = WSGIResponse({}, self.start_response)
        response.redirect('http://infrae.com')
        self.assertEqual(response.getStatus(), 302)
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '302 Moved Temporarily')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '0'),
             ('Location', 'http://infrae.com')])
        self.assertEqual(self.start_response.data, [])

    def test_insert_base(self):
        """Test inserting base: HTTPResponse compatiblity.
        """
        response = WSGIResponse({}, self.start_response)
        # We need to set content-type and base to trigger the
        # behavior, and have a full HTML page.
        response.setHeader('Content-Type', 'text/html;charset=utf-8')
        response.setBase('http://localhost/base')
        response.setBody('<html><head><title>Test</title></head><body>Test!</body></html>')

        response.startWSGIResponse()
        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '103'),
             ('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, [])
        self.assertEqual(
            response.body,
            '<html><head>\n<base href="http://localhost/base/" />\n'
            '<title>Test</title></head><body>Test!</body></html>')

    def test_no_insert_base_if_already_there(self):
        """Test inserting base: don't override existing one:
        HTTPResponse compatibility.
        """
        response = WSGIResponse({}, self.start_response)
        # We need to set content-type and base to trigger the
        # behavior, and have a full HTML page.
        response.setHeader('Content-Type', 'text/html;charset=utf-8')
        response.setBase('http://localhost/base')
        response.setBody(
            '<html><head><title>Test</title><base href="http://google.com/base" />'
            '</head><body>Test!</body></html>')

        response.startWSGIResponse()
        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '101'),
             ('Content-Type', 'text/html;charset=utf-8')])
        self.assertEqual(self.start_response.data, [])
        self.assertEqual(
            response.body,
            '<html><head><title>Test</title><base href="http://google.com/base" />'
            '</head><body>Test!</body></html>')

    def test_no_insert_base_if_not_text_html(self):
        """Test inserting base: only in text/html: HTTPResponse compatibility.
        """
        response = WSGIResponse({}, self.start_response)
        response.setHeader('Content-Type', 'text/text')
        response.setBase('http://localhost/base')
        response.setBody('<html><head><title>Test</title></head><body>Test!</body></html>')

        response.startWSGIResponse()
        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '63'),
             ('Content-Type', 'text/text')])
        self.assertEqual(self.start_response.data, [])
        self.assertEqual(
            response.body,
            '<html><head><title>Test</title></head><body>Test!</body></html>')

    def test_cookies(self):
        """Try to set a cookie.
        """
        response = WSGIResponse({}, self.start_response)
        response.setCookie('Silva', 'Best CMS Ever')
        response.setBody('<p>Silva</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Set-Cookie', 'Silva="Best%20CMS%20Ever"')])
        self.assertEqual(self.start_response.data, [])

    def test_cookies_expire(self):
        """Expire a cookie.
        """
        response = WSGIResponse({}, self.start_response)
        response.setCookie('Silva', 'Best CMS Ever')
        response.expireCookie('Silva')
        response.setBody('<p>Silva</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Set-Cookie', 'Silva="deleted"; Max-Age=0; Expires=Wed, 31-Dec-97 23:59:59 GMT')])
        self.assertEqual(self.start_response.data, [])

    def test_cookies_multiples(self):
        """Set more than one cookie.
        """
        response = WSGIResponse({}, self.start_response)
        response.setCookie('Silva', 'Best CMS Ever')
        response.setCookie('WSGI', 'Pluggable')
        response.setBody('<p>Silva</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Set-Cookie', 'Silva="Best%20CMS%20Ever"'),
             ('Set-Cookie', 'WSGI="Pluggable"')])
        self.assertEqual(self.start_response.data, [])

    def test_cookies_append(self):
        """Append a value at the end of a cookie.
        """
        response = WSGIResponse({}, self.start_response)
        response.setCookie('Silva', 'Best CMS Ever')
        response.appendCookie('Silva', 'Soon in 3D')
        response.setBody('<p>Silva</p>')
        response.startWSGIResponse()

        self.assertEqual(self.start_response.status, '200 OK')
        self.assertEqual(
            self.start_response.headers,
            [('Content-Length', '12'),
             ('Content-Type', 'text/html;charset=utf-8'),
             ('Set-Cookie', 'Silva="Best%20CMS%20Ever%3ASoon%20in%203D"')])
        self.assertEqual(self.start_response.data, [])

def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(ResponseTestCase))
    return suite



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from infrae.wsgi.interfaces import IRequest

from zope.component import getMultiAdapter
from zope.interface import implements
from ZODB.POSException import ConflictError

import Acquisition
import ExtensionClass


class MockWSGIStartResponse(object):
    """This mock object is used instead of the start_response callable
    of WSGI. It records any call to it.
    """

    __slots__ = ['status', '__headers', 'data']

    def __init__(self):
        self.status = None
        self.__headers = None
        self.data = []

    @apply
    def headers():
        def set(self, value):
            self.__headers = value
        def get(self):
            if self.__headers is None:
                return None
            headers = list(self.__headers)
            headers.sort(key=lambda t: t[0])
            return headers
        return property(get, set)

    def __call__(self, status, headers):
        assert self.status is None, "Status already set in mock response"
        assert self.headers is None, "Headers already set in mock response"
        self.status = status
        self.headers = headers
        return self.write

    def write(self, data):
        self.data.append(data)


class Mocker(object):

    def __init__(self):
        self.__called = []

    def __getattr__(self, method_name):
        if method_name.startswith('_'):
            raise AttributeError(method_name)
        def method(*args, **kwargs):
            self.mocker_call(method_name, args, kwargs)
        return method

    def mocker_call(self, method_name, args, kwargs):
        if not method_name.startswith('mocker_'):
            self.__called.append((method_name, args, kwargs))

    def mocker_called(self):
        called = list(self.__called)
        del self.__called[:]
        return called


class MockTransactionManager(Mocker):
    """Mock a Zope transaction manager.
    """

    def __init__(self):
        super(MockTransactionManager, self).__init__()
        self.__conflict = False

    def mocker_set_conflict(self, state):
        self.__conflict = state

    def commit(self):
        self.mocker_call('commit', (), {})
        if self.__conflict:
            if not isinstance(self.__conflict, bool):
                self.__conflict -= 1
            raise ConflictError()


class MockRoot(ExtensionClass.Base, Acquisition.Implicit):
    """Mockup Application.
    """

    def getPhysicalRoot(self):
        return self

    def getPhysicalPath(self):
        return ('',)

    def __bobo_traverse__(self, request):
        return self


class MockRequest(Mocker):
    implements(IRequest)

    def __init__(self, args=(), data={}, view=None, response=None, retry=0):
        super(MockRequest, self).__init__()
        self.response = response
        self.args = args
        self.environ = data.copy()
        self.other = {}
        self.steps = []
        self._steps = []
        self._hacked_path = 0
        self.__data = data
        self.__view = view
        self.__tries = retry

    def __getitem__(self, key):
        return self.__data[key]

    def __setitem__(self, key, value):
        self.__data[key] = value

    def get(self, key, default=None):
        return self.__data.get(key, default)

    def traverseName(self, path, name):
        return self.__view

    def supports_retry(self):
        return self.__tries != 0

    def retry(self):
        self.mocker_call('retry', (), {})
        assert self.__tries > 0, "Number of retries exceeded"
        self.__tries -= 1
        return self

    def query_plugin(self, context, iface):
        return getMultiAdapter((context, self), iface)

    def get_plugin(self, iface):
        return None


class MockApplication(object):
    """Mock an application.
    """

    def __init__(self, debug_exceptions=False):
        self.debug_exceptions = debug_exceptions
        self.transaction = MockTransactionManager()
        self.response = MockWSGIStartResponse()
        self.root = MockRoot()



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2012 Infrae. All rights reserved.
# See also LICENSE.txt

import urlparse
from urllib import quote

from five import grok
from zope.interface import Interface

import zExceptions

from infrae.wsgi.interfaces import IRequest, IVirtualHosting
from infrae.wsgi.utils import split_path_info, traverse


class VirtualHosting(grok.MultiAdapter):
    grok.adapts(Interface, IRequest)
    grok.provides(IVirtualHosting)
    grok.implements(IVirtualHosting)

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.root = None

    def rewrite_url(self, base_url, original_url):
        base = (None, None)
        if base_url:
            base = urlparse.urlparse(base_url)[:2]
        return urlparse.urlunparse(
            base + urlparse.urlparse(original_url)[2:])

    def __call__(self, method, path):
        root = self.context
        virtual_host = self.request.environ.get('HTTP_X_VHM_URL')
        if virtual_host:
            url = urlparse.urlparse(virtual_host)

            # Step 1. Set Server Name
            hostname = url.netloc
            if ':' in hostname:
                hostname, port = url.netloc.split(':', 1)
            else:
                port = '80'
                if url.scheme == 'https':
                    port = '443'
            self.request.setServerURL(url.scheme, hostname, port)

            # Step 2. Consume extra path
            host_path = split_path_info(url.path)
            if host_path:
                for piece in host_path:
                    if path:
                        taken_piece = path.pop()
                        if taken_piece == piece:
                            continue
                    raise zExceptions.BadRequest(
                        u'This URL is not in the virtual host.')

            # Step 3. Traverse to the virtual root
            virtual_path = split_path_info(
                self.request.environ.get('HTTP_X_VHM_PATH'))
            if virtual_path:
                root = traverse(virtual_path, self.context, self.request)
                self.root = root

            # Step 4, in case of path manipulation, set virtual root
            if virtual_path or host_path:
                self.request.setVirtualRoot(host_path)

            # Update Actual URL
            self.request['ACTUAL_URL'] = (
                self.request['URL'] + quote(self.request['PATH_INFO']))

        return root, method, path



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

import unittest

import infrae.wsgi
from infrae.wsgi.testing import BrowserLayer, Browser
from infrae.testing import get_event_names
from zope.security.management import queryInteraction


class WSGILayer(BrowserLayer):
    default_users = {'admin': ['Manager']}


FunctionalLayer = WSGILayer(infrae.wsgi)


class FunctionalTestCase(unittest.TestCase):
    """Functional testing.
    """
    layer = FunctionalLayer

    def setUp(self):
        self.browser = Browser()
        self.browser.handleErrors = True
        self.browser.raiseHttpErrors = False
        get_event_names()       # Clear setup events

    def test_default_view(self):
        self.browser.open('http://localhost')
        self.assertEqual(self.browser.title, 'Zope QuickStart')
        self.assertEqual(self.browser.status, '200 OK')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationAfterTraversal',
             'PublicationAfterRender',
             'PublicationBeforeCommit',
             'PublicationSuccess',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)

    def test_notfound_view(self):
        self.browser.open('http://localhost/nowhere')
        self.assertEqual(self.browser.status, '404 Not Found')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationBeforeError',
             'PublicationAfterRender',
             'PublicationBeforeAbort',
             'PublicationFailure',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)

    def test_debug_view(self):
        # You must be authenticated to access the debug view.
        self.browser.open('http://localhost/debugzope.html')
        self.assertEqual(self.browser.status, '401 Unauthorized')
        self.browser.addHeader('Authorization', 'Basic admin:admin')
        self.browser.reload()
        self.assertEqual(self.browser.status, '200 OK')

    def test_log_view(self):
        # You must be authenticated to access the log view.
        self.browser.open('http://localhost/errorlog.html')
        self.assertEqual(self.browser.status, '401 Unauthorized')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationBeforeCommit',
             'PublicationSuccess',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)

        self.browser.addHeader('Authorization', 'Basic admin:admin')
        self.browser.reload()
        self.assertEqual(self.browser.status, '200 OK')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationAfterTraversal',
             'PublicationAfterRender',
             'PublicationBeforeCommit',
             'PublicationSuccess',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)

    def test_abort(self):
        self.browser.open('http://localhost/abort.html')
        self.assertEqual(self.browser.status, '400 Bad Request')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationAfterTraversal',
             'PublicationBeforeAbort',
             'PublicationFailure',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)

    def test_conflict(self):
        self.browser.open('http://localhost/conflict.html')
        self.assertEqual(self.browser.status, '503 Service Unavailable')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationAfterTraversal',
             'PublicationBeforeAbort',
             'PublicationFailure'] * 4 +
            ['EndRequestEvent'] * 4)
        self.assertIs(queryInteraction(), None)

    def test_failed_conflict(self):
        self.browser.open('http://localhost/failed_conflict.html')
        self.assertEqual(self.browser.status, '500 Internal Server Error')
        self.assertEqual(
            get_event_names(),
            ['PublicationStart',
             'PublicationAfterTraversal',
             'PublicationBeforeAbort',
             'PublicationFailure',
             'PublicationStart',
             'PublicationAfterTraversal',
             'PublicationBeforeError',
             'PublicationAfterRender',
             'PublicationBeforeAbort',
             'PublicationFailure',
             'EndRequestEvent',
             'EndRequestEvent'])
        self.assertIs(queryInteraction(), None)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(FunctionalTestCase))
    return suite



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from OFS.interfaces import IObjectManager
from zExceptions import BadRequest
from urllib import quote


def reconstruct_url_from_environ(environ):
    """Reconstruct an URL from the WSGI environ.
    """
    # This code is taken from the PEP333
    url = environ['wsgi.url_scheme']+'://'

    if environ.get('HTTP_HOST'):
        url += environ['HTTP_HOST']
    else:
        url += environ['SERVER_NAME']

        if environ['wsgi.url_scheme'] == 'https':
            if environ['SERVER_PORT'] != '443':
               url += ':' + environ['SERVER_PORT']
        else:
            if environ['SERVER_PORT'] != '80':
               url += ':' + environ['SERVER_PORT']

    url += quote(environ.get('SCRIPT_NAME', ''))
    url += quote(environ.get('PATH_INFO', ''))
    if environ.get('QUERY_STRING'):
        url += '?' + environ['QUERY_STRING']
    return url


def split_path_info(path):
    """Split a path from a string as a list of components, removing
    what is useless.
    """
    result = []
    if not path:
        return result
    for item in path.split('/'):
        if item in ('REQUEST', 'aq_self', 'aq_base'):
            raise BadRequest(u"Invalid path %s" % path)
        if not item or item=='.':
            continue
        elif item == '..':
            if not len(result):
                raise BadRequest(u"Invalid path %s" % path)
            del result[-1]
        else:
            result.append(item)
    return result


def traverse(path, content, request=None):
    """Simple helper to traverse path from content, following only
    Zope 2 containers API.
    """
    for piece in path:
        if not IObjectManager.providedBy(content):
            raise BadRequest(
                u'Invalid path, not a container /%s.' % '/'.join(path))
        child = content._getOb(piece, None)
        if child is None:
            raise BadRequest(
                u'Invalid path, content not found /%s' % '/'.join(path))
        if request is not None:
            hook = getattr(child, '__before_publishing_traverse__', None)
            if hook is not None:
                hook(child, request)
            request['PARENTS'].append(child)
        content = child
    return content



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

import threading
import logging
import pdb
import sys

from Zope2 import startup
from infrae.wsgi.publisher import WSGIApplication
from infrae.wsgi.log import reporter
from zope.event import notify
from zope.processlifetime import ProcessStarting
import App.config
import Zope2

logger = logging.getLogger('infrae.wsgi')
bootstrap_lock = threading.Lock()
bootstrap_done = False


def configure_zope(config_filename, debug_mode=False):
    """Read zope.conf with zdaemon^Wzcrap.
    """
    from Zope2.Startup import options, handlers
    import AccessControl

    del sys.argv[1:]
    opts = options.ZopeOptions()
    opts.configfile = config_filename
    opts.realize(raise_getopt_errs=0)

    handlers.handleConfig(opts.configroot, opts.confighandlers)
    AccessControl.setImplementation(
        opts.configroot.security_policy_implementation)
    AccessControl.setDefaultBehaviors(
        not opts.configroot.skip_ownership_checking,
        not opts.configroot.skip_authentication_checking,
        opts.configroot.verbose_security)
    App.config.setConfiguration(opts.configroot)
    set_zope_debug_mode(debug_mode)


def set_zope_debug_mode(debug_mode):
    """Set the Zope debug mode to the given value.
    """
    config = App.config.getConfiguration()
    config.debug_mode = debug_mode and 1 or 0
    import Globals
    Globals.DevelopmentMode = config.debug_mode


def mount_all_databases():
    """Call this to mount all available Zope databases.
    """
    config = App.config.getConfiguration()
    connection = Zope2.DB.open()
    root = connection.root()['Application']
    for path, name in config.dbtab.listMountPaths():
        if path != '/':
            try:
                # Get the database.
                # If it doesn't exist yet, it will be created.
                config.dbtab.getDatabase(name=name)
                root.unrestrictedTraverse(path)
                logger.info("Mounted %s on %s" % (name, path))
            except AttributeError:
                pass
    connection.close()


def boot_zope(config_filename, debug_mode=False):
    """Boot Zope.
    """
    global bootstrap_done
    bootstrap_lock.acquire()
    try:
        if bootstrap_done:
            logger.info("Zope already configured, skipping configuration")
            return
        configure_zope(config_filename, debug_mode)
        logger.info("Zope configured")

        try:
            startup()
            mount_all_databases()
        except Exception:
            if debug_mode:
                # If debug_mode is on, debug possible starting errors.
                print "%s:" % sys.exc_info()[0]
                print sys.exc_info()[1]
                pdb.post_mortem(sys.exc_info()[2])
            raise

        # Some products / Zope code reset the debug mode. Re-set it again.
        set_zope_debug_mode(debug_mode)

        # Notify start of application
        notify(ProcessStarting())
        bootstrap_done = True
        logger.info("Zope started")
    finally:
        bootstrap_lock.release()

def parse_raven_config(options):
    config = configuration_dict(options, 'raven', sep=".")
    for key in ['exclude_paths', 'include_paths', 'processors']:
        if key in config:
            config[key] = configuration_list(config[key])
    if config:
        from .ravenplugin import RavenLoggingPlugin
        reporter.subscribe_to_errors(RavenLoggingPlugin(config))

def zope2_application_factory(global_conf, zope_conf, **options):
    """Build a Zope2 WSGI application.
    """
    parse_raven_config(options)

    debug_mode = options.get('debug_mode', 'off') == 'on'
    zope_workers = configuration_int(options, 'zope_workers')
    boot_zope(zope_conf, debug_mode)

    # By default debug mode does not render exceptions as error pages
    # but allow them to propagate to outer wsgi middlewares
    # (e.g. z3c.evalexception). Setting this to 'off' will disable the
    # propagation
    debug_exceptions = (debug_mode and
                        options.get('debug_exceptions', 'on') == 'on')

    ignore_errors = configuration_list(options, 'ignore_errors')
    if ignore_errors:
        reporter.ignore_errors(ignore_errors)
    show_errors = configuration_list(options, 'show_errors')
    if show_errors:
        reporter.show_errors(show_errors)

    return WSGIApplication(
        Zope2.bobo_application,
        Zope2.zpublisher_transactions_manager,
        debug_mode,
        debug_exceptions,
        zope_workers)

# Utilities to read configuration options

def configuration_dict(options, key, sep="."):
    key_sep = key + sep
    result = {}
    for key, value in options.iteritems():
        if key.startswith(key_sep):
            prefix, result_key = key.split(sep, 1)
            result[result_key] = value
    return result

def configuration_list(options, key, default=''):
    return filter(lambda value: value,
                  map(lambda value: value.strip(),
                      options.get(key, default).split(',')))

def configuration_int(options, key, default='4'):
    value = options.get(key, default)
    try:
        return int(value)
    except:
        return int(default)



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from cStringIO import StringIO
from tempfile import TemporaryFile
from urllib import quote
import socket
import threading

from zope.cachedescriptors.property import Lazy

from AccessControl.SecurityManagement import noSecurityManager
from Acquisition.interfaces import IAcquirer
from ZPublisher.BaseRequest import exec_callables, RequestContainer
from ZPublisher.Publish import Retry
from ZPublisher.Request import Request
from ZPublisher.mapply import mapply
from ZODB.POSException import ConflictError
from zope.component import queryMultiAdapter, getMultiAdapter
from zope.event import notify
from zope.interface import implements
from zope.site.hooks import getSite
from zope.publisher.interfaces.browser import IBrowserPage, IBrowserPublisher
from zope.security.management import newInteraction, endInteraction
import Zope2
import zExceptions

from infrae.wsgi import interfaces
from infrae.wsgi.errors import DefaultError
from infrae.wsgi.response import WSGIResponse, AbortPublication
from infrae.wsgi.log import logger, log_last_error, ErrorSupplement
from infrae.wsgi.utils import reconstruct_url_from_environ
from infrae.wsgi.utils import split_path_info

CHUNK_SIZE = 1<<16              # 64K


def call_object(obj, args, request):
    return apply(obj, args)


def missing_name(name, request):
    if name == 'self':
        return request['PARENTS'][0]
    raise zExceptions.BadRequest(name)


def dont_publish_class(klass, request):
    raise zExceptions.Forbidden("class %s" % klass.__name__)


class WSGIRequest(Request):
    """A WSGIRequest have a default skin.
    """
    implements(interfaces.IRequest)

    def __init__(self, *args, **kwargs):
        Request.__init__(self, *args, **kwargs)
        self.__plugins = {interfaces.IRequest.__identifier__: self}

    def query_plugin(self, context, iface):
        plugin = getMultiAdapter((context, self), iface)
        self.__plugins[iface.__identifier__] = plugin
        return plugin

    def get_plugin(self, iface):
        return self.__plugins.get(iface.__identifier__)


class WSGIResult(object):
    """Iterator to wrap Zope result, in order to commit/abort the
    transaction at the end of the iterator iteration, and to close the
    request at the end.
    """

    def __init__(self, request, publisher, data, cleanup=None):
        self.request = request
        self.publisher = publisher
        self.cleanup = cleanup
        self.__next = iter(data).next
        self.__failed = False

    @Lazy
    def url(self):
        return reconstruct_url_from_environ(self.request.environ)

    def next(self):
        try:
            return self.__next()
        except StopIteration:
            raise
        except Exception:
            logger.exception(
                u"Unexpected error in the WSGI stack for '%s'", self.url)
            self.__failed = True
            raise

    def __iter__(self):
        return self

    def close(self):
        try:
            if self.__failed:
                logger.error(
                    "An error happened in the WSGI stack "
                    "while iterating the result for the url '%s'", self.url)
                self.publisher.abort()
            else:
                self.publisher.finish()
        except ConflictError:
            self.publisher.abort()
            raise
        finally:
            # Always cleanup
            if self.cleanup is not None:
                self.cleanup()


def safe_callback(publisher, func, *args, **kwargs):
    try:
        func(*args, **kwargs)
    except Exception:
        log_last_error(publisher.request, publisher.response)


class WSGIPublication(object):
    """Publish a request through WSGI.
    """

    def __init__(self, app, request, response):
        self.app = app
        self.request = request
        self.response = response
        self.data_sent = False
        self.publication_done = False

    def start(self):
        """Start the publication process.
        """
        notify(interfaces.PublicationStart(self.request))
        newInteraction()
        noSecurityManager()
        self.app.transaction.begin()
        self.request.processInputs()

    def commit(self):
        """Commit results of the publication.
        """
        safe_callback(
            self,
            notify,
            interfaces.PublicationBeforeCommit(self.request))
        self.app.transaction.commit()
        safe_callback(
            self,
            notify,
            interfaces.PublicationSuccess(self.request))

    def abort(self):
        """Abort the current publication process.
        """
        safe_callback(
            self,
            notify,
            interfaces.PublicationBeforeAbort(self.request, None, False))
        self.app.transaction.abort()
        safe_callback(
            self,
            notify,
            interfaces.PublicationFailure(self.request, None, False))

    def finish(self):
        """End the publication process, by either committing the
        transaction or aborting it.
        """
        if self.publication_done:
            return
        if self.response.getStatus() < 404:
            # We want to commit the transaction in case of redirects
            # and unauthorized.
            self.commit()
        else:
            self.abort()
        self.publication_done = True

    def result(self):
        """Return the result of the response, and commit if we don't
        plan to stream data/or had stream data.
        """
        self.data_sent, data = self.response.getWSGIResponse()
        if not self.data_sent:
            # If we didn't send any data yet, commit now: if the commit
            # fail, we can retry the request since we didn't send
            # anything.
            self.finish()
        return data

    def render_error(self, error, last_known_obj):
        """Render and log an error.
        """
        if IBrowserPage.providedBy(last_known_obj):
            #of the last obj is a view, use it's context (which should be
            # an IAcquirer)
            last_known_obj = last_known_obj.context
        if not IAcquirer.providedBy(last_known_obj):
            last_known_site = getSite()
            if last_known_site is not None:
                last_known_obj = last_known_site
        context = DefaultError(error)
        if IAcquirer.providedBy(last_known_obj):
            context = context.__of__(last_known_obj)
        error_view = queryMultiAdapter(
            (context, self.request), name='error.html')
        if IBrowserPublisher.providedBy(error_view):
            error_view, error_path = error_view.browserDefault(self.request)
            if error_path:
                raise NotImplementedError(
                    u'Error browserDefault retuned an path. '
                    u'This is not implemented.')

        if error_view is not None:
            notify(interfaces.PublicationBeforeError(
                    self.request, last_known_obj))
            try:
                error_result = error_view()
                if error_result is not None:
                    self.response.setBody(error_result)

                notify(interfaces.PublicationAfterRender(
                        self.request, error_view))

            except Exception:
                log_last_error(
                    self.request, self.response, obj=last_known_obj,
                    extra=u"Error while rendering error message")
                self.response.setStatus(500)
                self.response.setBody(ERROR_WHILE_RENDERING_ERROR_TEMPLATE)
        else:
            logger.error('No action defined for last exception')
            self.response.setStatus(500)
            self.response.setBody(DEFAULT_ERROR_TEMPLATE)

    def should_render_errors(self):
        """Return True if the error pages should be rendered.
        """
        return self.request.environ.get(
            'wsgi.handleErrors',
            not self.app.debug_exceptions)

    def get_application_root(self):
        """Return a new Zope root for this request.
        """
        # This create the connection to the ZODB, wrap it in the
        # request.  After this, request.close must always be called.
        root = self.app.root.__bobo_traverse__(self.request)
        return root.__of__(RequestContainer(REQUEST=self.request))

    def get_path_and_method(self, root):
        """Return the path and the method of this request.
        """
        request = self.request
        path_info = request['PATH_INFO']

        # Inspect path
        path = list(reversed(split_path_info(path_info)))
        request['ACTUAL_URL'] = request['URL'] + quote(path_info)
        request['TraversalRequestNameStack'] = request.path = path
        request['PARENTS'] = [root]

        # Method
        method = request.get('REQUEST_METHOD', 'GET').upper()
        if method in ['GET', 'POST']:
            # index_html is still the default method, only any object can
            # override it by implementing its own __browser_default__ method
            method = 'index_html'

        if not path and not method:
            raise zExceptions.BadRequest(request['URL'])

        return method, path

    def publish(self):
        """Publish the request into the response.
        """
        self.start()

        # First check for "cancel" redirect ZMI-o-hardcoded thing:
        submit = self.request.get('SUBMIT', None)
        if submit is not None:
            if submit.strip().lower() == 'cancel':
                cancel = self.request.get('CANCEL_ACTION','')
                if cancel:
                    raise zExceptions.Redirect(cancel)

        # Get the path, method and root
        root = self.get_application_root()
        method, path = self.get_path_and_method(root)

        # Do some optional virtual hosting
        vhm = self.request.query_plugin(
            root, interfaces.IVirtualHosting)
        root, method, path = vhm(method, path)

        # Zope 2 style post traverser hooks
        self.request._post_traverse = post_traverse = []

        # Get object to publish/render
        traverser = self.request.query_plugin(root, interfaces.ITraverser)
        content = traverser(method, path)
        __traceback_supplement__ = (ErrorSupplement, content)

        # Zope 2 style post traverser hooks
        del self.request._post_traverse

        # Run authentication
        authenticator = self.request.query_plugin(
            content, interfaces.IAuthenticator)
        authenticator(Zope2.zpublisher_validated_hook)

        # Run Zope 2 style post traversal hooks
        if post_traverse:
            result = exec_callables(post_traverse)
            if result is not None:
                content = result

        notify(interfaces.PublicationAfterTraversal(self.request, content))

        # Render the content into the response
        self.app.transaction.recordMetaData(content, self.request)
        result = mapply(
            content, self.request.args, self.request,
            call_object, 1, missing_name, dont_publish_class,
            self.request, bind=1)

        if result is not None:
            self.response.setBody(result)

        notify(interfaces.PublicationAfterRender(self.request, content))

        return self.result()

    def publish_and_render_errors(self):
        """Publish the request, manage errors.
        """
        def last_content():
            content = self.request.get('PUBLISHED')
            if content is None:
                parents = self.request.get('PARENTS')
                if parents:
                    return parents[0]
            return content

        try:
            return self.publish()
        except (ConflictError, Retry, AbortPublication):
            # Conflict are managed at an higher level
            raise
        except zExceptions.Unauthorized:
            # Manage unauthorized
            log_last_error(self.request, self.response, last_content())
            self.response.setStatus(401)
            try:
                # To be compatible with PAS
                self.response._unauthorized()
            except Exception:
                # The _unauthorised handler using PAS failed
                log_last_error(
                    self.request, self.response, obj=last_content(),
                    extra="Error while processing the unauthorized PAS handler")
                self.response.setStatus(401)
                self.response.setBody("")
                self.response.setHeader(
                    'WWW-Authenticate', 'basic realm="%s"' % self.response.realm)
        except zExceptions.Redirect as error:
            # Redirect
            self.response.redirect(str(error))
        except Exception as error:
            content = last_content()
            log_last_error(self.request, self.response, content)
            if not self.should_render_errors():
                # If we don't handle errors, reraise the exception.
                raise

            self.render_error(error, content)

        # Return the result of the response
        return self.result()

    def publish_and_retry(self):
        """Publish the request into the response and retry if it
        fails.
        """
        try:
            data = self.publish_and_render_errors()
        except (ConflictError, Retry):
            self.abort()
            self.publication_done = True
            if self.request.supports_retry() and not self.data_sent:
                # If can still retry, and didn't send any data yet, do it.
                logger.info('Conflict, retrying request %s' % (
                        reconstruct_url_from_environ(self.request.environ)))
                endInteraction()
                request = self.request.retry()
                try:
                    publication = self.__class__(
                        self.app, request, self.response)
                    data = publication.publish_and_retry()
                    self.publication_done = publication.publication_done
                finally:
                    request.close()
            else:
                # Otherwise, just render a plain error.
                logger.error('Conflict error for request %s' % (
                        reconstruct_url_from_environ(self.request.environ)))
                self.response.setStatus(503)
                self.response.setBody(RETRY_FAIL_ERROR_TEMPLATE)
                data = self.result()
        return data

    def __call__(self, cleanup=None):
        """Publish the request and send the result via an iterator.
        """
        try:
            data = self.publish_and_retry()
            self.response.startWSGIResponse()
            return WSGIResult(self.request, self, data, cleanup)
        except Exception:
            # In case of exception we didn't catch (like
            # AbortPublication), abort all the current transaction.
            self.abort()
            if cleanup is not None:
                cleanup()
            raise


class WSGIApplication(object):
    """Zope WSGI application.
    """

    def __init__(self, root, transaction,
                 debug_mode=False, debug_exceptions=True,
                 concurrency=4):
        self.root = root
        self.transaction = transaction
        self.memory_maxsize = 2 << 20
        self.debug_mode = debug_mode
        self.debug_exceptions = debug_exceptions
        self.concurrency = threading.Semaphore(concurrency)

    def save_input(self, environ):
        """We want to save the request input in order to be able to
        retry a request: Zope need to be able to do .seek(0) on
        wsgi_input.
        """
        original_input = environ.get('wsgi.input')

        if original_input is not None:
            environ_length = environ.get('CONTENT_LENGTH')
            length = int(environ_length) if environ_length else 0
            if length > self.memory_maxsize:
                new_input = environ['wsgi.input'] = TemporaryFile('w+b')
            else:
                new_input = environ['wsgi.input'] = StringIO()
            to_read = length

            try:
                while to_read:
                    if to_read <= CHUNK_SIZE:
                        data = original_input.read(to_read)
                        to_read = 0
                    else:
                        data = original_input.read(CHUNK_SIZE)
                        to_read -= CHUNK_SIZE
                    new_input.write(data)
            except (socket.error, IOError):
                raise AbortPublication(started=False)

            new_input.seek(0)
            environ['wsgi.input'] = new_input

    def __call__(self, environ, start_response):
        """WSGI entry point.
        """
        try:
            self.save_input(environ)
            response = WSGIResponse(environ, start_response, self.debug_mode)
            request = WSGIRequest(environ['wsgi.input'], environ, response)
            publication = WSGIPublication(self, request, response)

            self.concurrency.acquire()
            def cleanup(context=None):
                request.close()
                endInteraction()
                noSecurityManager()
                self.concurrency.release()

            return publication(cleanup)
        except AbortPublication, error:
            if not error.response_started:
                msg = 'Socket error'
                start_response('400 Bad Request', [
                        ('Content-Type', 'text/plain'),
                        ('Content-Length', str(len(msg))),
                        ])
                return [msg]
            # Return nothing otherwise
            return []



ERROR_WHILE_RENDERING_ERROR_TEMPLATE = u"""
<html>
  <head>
    <title>Error</title>
  </head>
  <body>
     <h1>An error happened while rendering an error message</h1>
     <p>Sorry for the inconvenience.</p>
     <p>Please check the server logs for more information.</p>
  </body>
</html>
"""

DEFAULT_ERROR_TEMPLATE = u"""
<html>
  <head>
    <title>Error</title>
  </head>
  <body>
     <h1>An error happened</h1>
     <p>Sorry for the inconvenience.</p>
     <p>Please check the server logs for more information.</p>
  </body>
</html>
"""

RETRY_FAIL_ERROR_TEMPLATE = u"""
<html>
  <head>
    <title>Service temporarily unavailable</title>
  </head>
  <body>
     <h1>Your request cannot be processed at the moment</h1>
     <p>Please retry later.</p>
  </body>
</html>
"""





# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from cgi import escape
from datetime import datetime
import collections
import logging
import sys

from zExceptions.ExceptionFormatter import format_exception
from zope.browser.interfaces import IView
from zope.interface import Interface
from five import grok

from infrae.wsgi.utils import reconstruct_url_from_environ

logger = logging.getLogger('infrae.wsgi')


def object_name(obj):
    """Return an object name
    """
    return '%s.%s' % (obj.__class__.__module__, obj.__class__.__name__)


def object_path(obj):
    """Return an object path in the Zope database.
    """
    try:
        if hasattr(obj, 'getPhysicalPath'):
            return '/'.join(obj.getPhysicalPath())
    except:
        pass
    return 'n/a'


def log_invalid_response_data(data, environ):
    """Log an invalid response type from application. Data sent must
    always be a string (unicode strings are accepted), if it is not an
    IResult or IStreamIterator object (those must behave correctly).
    """
    logger.error(
        "Invalid response data of type %s for url '%s'" %
        (object_name(data), reconstruct_url_from_environ(environ)))


class ErrorLogView(grok.View):
    grok.context(Interface)
    grok.name('errorlog.html')
    grok.require('zope2.ViewManagementScreens')

    def update(self):
        if 'ignore_errors_update' in self.request.form:
            reporter.ignored_errors = self.request.form.get(
                'ignore_errors', [])

        self.all_errors = reporter.all_ignored_errors
        self.ignored_errors = reporter.ignored_errors
        self.errors = reporter.get_last_errors()
        self.debug_mode = self.request.response.debug_mode


class ErrorSupplement(object):
    """Add more information about an error on a view in a traceback.
    """

    def __init__(self, cls):
        self.context = cls
        if IView.providedBy(cls):
            self.context = cls.context
        self.cls = cls

    def getInfo(self, as_html=0):
        info = list()
        info.append((u'Published class', object_name(self.cls),))
        info.append((u'Object path', object_path(self.context),))
        info.append(
            (u'Object type', getattr(self.context, 'meta_type', u'n/a',)))
        if not as_html:
            return '   - ' + '\n   - '.join(map(lambda x: '%s: %s' % x, info))

        return u'<p>Extra information:<br /><li>%s</li></p>' % ''.join(map(
                lambda x: u'<li><b>%s</b>: %s</li>' % (
                    escape(str(x[0])), escape(str(x[1]))),
                info))


class ErrorReporter(object):
    """Utility to help error reporting.
    """
    all_ignored_errors = [
        'NotFound', 'Redirect',
        'Unauthorized', 'Forbidden',
        'BadRequest', 'BrokenReferenceError']

    def __init__(self, subscribed=[]):
        self.__last_errors = collections.deque([], 25)
        self.__ignore_errors = set(self.all_ignored_errors)
        self.__subscribed = list(subscribed) + [self._subscriber]

    def _subscriber(self, request, response, obj, exc_info, exc_text, extra):
        log_entry = ['\n']

        if extra is not None:
            log_entry.append(extra + '\n')

        if obj is not None:
            log_entry.append('Object class: %s\n' % object_name(obj))
            log_entry.append('Object path: %s\n' % object_path(obj))

        def log_request_info(title, key):
            value = request.get(key, 'n/a') or 'n/a'
            log_entry.append('%s: %s\n' % (title, value))

        log_request_info('Request URL', 'URL')
        log_request_info('Request method', 'method')
        log_request_info('Query string', 'QUERY_STRING')
        log_request_info('User', 'AUTHENTICATED_USER')
        log_request_info('User-agent', 'HTTP_USER_AGENT')
        log_request_info('Refer', 'HTTP_REFERER')

        log_entry.extend(exc_text)

        # Save error.
        report = ''.join(log_entry)
        logger.error(report)
        self.__last_errors.append(
            {'url': request['URL'], 'report': report, 'time': datetime.now()})

    def subscribe_to_errors(self, handler):
        """Subscribe when an error happens.
        """
        self.__subscribed.append(handler)

    def show_errors(self, errors):
        """Show the given errors.
        """
        self.__ignore_errors -= set(errors)

    def ignore_errors(self, errors):
        """Ignore the given errors.
        """
        self.__ignore_errors += set(errors)

    @apply
    def ignored_errors():
        """Access currently ignored errors.
        """

        def getter(self):
            return list(self.__ignore_errors)

        def setter(self, errors):
            self.__ignore_errors = set(errors)

        return property(getter, setter)

    def is_loggable(self, error):
        """Tells you if this error is loggable.
        """
        error_name = error.__class__.__name__
        return error_name not in self.__ignore_errors

    def log_last_error(self, request, response, obj=None, extra=None):
        """Build an error report and log the last available error.
        """
        error_type, error_value, traceback = exc_info = sys.exc_info()
        try:
            if ((not response.debug_mode) and
                (not self.is_loggable(error_value))):
                return

            exc_text = format_exception(error_type, error_value, traceback)

            for plugin in self.__subscribed:
                plugin(request, response, obj, exc_info, exc_text, extra)
        finally:
            del traceback

    def get_last_errors(self):
        """Return all last errors.
        """
        errors = list(self.__last_errors)
        errors.reverse()
        return errors


reporter = ErrorReporter()


def log_last_error(request, response, obj=None, extra=None):
    """Log the last triggered error.
    """
    reporter.log_last_error(request, response, obj=obj, extra=extra)



# -*- coding: utf-8 -*-
# Copyright (c) 2012  Infrae. All rights reserved.
# See also LICENSE.txt

from five import grok
from zope.interface import Interface

import thread
import sys
import traceback
import time
from cStringIO import StringIO

def dump_threads():
    """Dump running threads

    Returns a string with the tracebacks.
    """

    frames = sys._current_frames()
    this_thread_id = thread.get_ident()
    now = time.strftime("%Y-%m-%d %H:%M:%S")
    res = ["Threads traceback dump at %s\n" % now]
    for thread_id, frame in frames.iteritems():
        if thread_id == this_thread_id:
            continue

        # Find request in frame
        reqinfo = ''
        f = frame
        while f is not None:
            co = f.f_code

            if co.co_name == 'publish':
                if co.co_filename.endswith('/infrae/wsgi/publisher.py'):
                    request = f.f_locals.get('request')
                    if request is not None:
                        reqinfo += (request.get('REQUEST_METHOD', '') + ' ' +
                                   request.get('PATH_INFO', ''))
                        qs = request.get('QUERY_STRING')
                        if qs:
                            reqinfo += '?'+qs
                    break
            f = f.f_back
        if reqinfo:
            reqinfo = " (%s)" % reqinfo

        output = StringIO()
        traceback.print_stack(frame, file=output)
        res.append("Thread %s%s:\n%s" %
            (thread_id, reqinfo, output.getvalue()))

    frames = None
    res.append("End of dump")
    return '\n'.join(res)


class DebugThreads(grok.View):
    grok.context(Interface)
    grok.name('debugzope.html')
    grok.require('zope2.ViewManagementScreens')

    def render(self):
        return u'<html><body><pre>%s</pre></body></html>' % dump_threads()


try:
    import webob.dec
except ImportError:
    def make_middleware(app, global_config, **config):
        return app
else:
    class DebugThreadsMiddleware(object):
        """Display the threads status. Warning: if enabled, this
        can be accessed without any authentication.
        """

        def __init__(self, app, name="++threads++"):
            self.app = app
            self.name = name

        @webob.dec.wsgify
        def __call__(self, request):
            if request.path_info.startswith('/' + self.name):
                threads = dump_threads()
                return '<html><body><pre>%s</pre></body></html>' % threads
            return request.get_response(self.app)

    def make_middleware(app, global_config, **config):
        name = config.get('name', '++threads++')
        return DebugThreadsMiddleware(app, name=name)




# -*- coding: utf-8 -*-
# Copyright (c) 2011-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from five import grok
from zope.component import queryMultiAdapter
from zope.interface import Interface
from zope.publisher.interfaces import NotFound
from zope.publisher.interfaces.browser import IBrowserPublisher

from AccessControl.ZopeSecurityPolicy import getRoles
from ZPublisher.BaseRequest import UNSPECIFIED_ROLES, DefaultPublishTraverse
from ZPublisher.BaseRequest import quote
from OFS.interfaces import IObjectManager
import zExceptions

from infrae.wsgi.interfaces import IRequest, ITraverser


class ObjectManagerPublishTraverse(DefaultPublishTraverse, grok.MultiAdapter):
    """The default Zope 3 folder traverser is register in Zope
    2.13.13. We register the Zope 2 one for Zope 2 folder, to prevent
    a broken ZMI.
    """
    grok.adapts(IObjectManager, IRequest)
    grok.provides(IBrowserPublisher)


def query_adapter(content, request, iface):
    """Query an adapter if needed.
    """
    if iface.providedBy(content):
        return content
    adapter = queryMultiAdapter((content, request), iface)
    if adapter is None:
        return DefaultPublishTraverse(content, request)
    return adapter


class Traverser(grok.MultiAdapter):
    grok.adapts(Interface, IRequest)
    grok.provides(ITraverser)
    grok.implements(ITraverser)

    def __init__(self, context, request):
        self.request = request
        self.context = context
        self.roles = getRoles(None, None, context, UNSPECIFIED_ROLES)

    def __call__(self, method, path):
        request = self.request
        content = self.context
        parents = request['PARENTS']
        request['TraversalRequestNameStack'] = path
        # Steps records the traversed path (used for some URL computation)
        steps = request.steps
        _steps = request._steps = map(quote, steps)
        entry_name = ''

        try:
            # We build parents in the wrong order, so we
            # need to make sure we reverse it when we're done.
            while 1:
                bpth = getattr(content, '__before_publishing_traverse__', None)
                if bpth is not None:
                    bpth(content, request)

                path = request.path = request['TraversalRequestNameStack']
                # Consume one item in the path.
                if path:
                    entry_name = path.pop()
                else:
                    # End of path, look for a view using IBrowserPublisher

                    next_content, default_path = query_adapter(
                        content, request, IBrowserPublisher).browserDefault(
                        request)

                    if next_content is not content:
                        self.roles = getRoles(
                            content, None, next_content, self.roles)
                        content = next_content

                    if default_path:
                        request._hacked_path=1
                        if len(default_path) > 1:
                            path = list(default_path)
                            method = path.pop()
                            request['TraversalRequestNameStack'] = path
                            continue
                        else:
                            entry_name = default_path[0]
                    elif (method and hasattr(content, method)
                          and entry_name != method
                          and getattr(content, method) is not None):
                        request._hacked_path=1
                        entry_name = method
                        method = 'index_html'
                    else:
                        break

                step = quote(entry_name)
                _steps.append(step)
                steps.append(entry_name)
                request['URL'] = '/'.join((request['URL'], step))

                try:
                    next_content = request.traverseName(content, entry_name)
                except (KeyError, AttributeError, NotFound):
                    raise zExceptions.NotFound(request['URL'])

                if (hasattr(content, '__bobo_traverse__') or
                    hasattr(content, entry_name)):
                    check_name = entry_name
                else:
                    check_name = None

                self.roles = getRoles(
                    content, check_name, next_content, self.roles)
                content = next_content

                parents.append(content)
        finally:
            parents.reverse()

        # Hacked path is for ZMI, where the real path is different
        # than the URL, and a base tag must be hacked in the body
        # response.
        if request._hacked_path:
            url = request['URL']
            index = url.rfind('/')
            if index > 0:
                request.response.setBase(url[:index])


        # content is PUBLISHED, or the first in parents (since it have
        # been reversed).
        request['PUBLISHED'] = parents.pop(0)

        # If content as a call method, get the roles defined on it, it
        # will be called.
        if hasattr(content, '__call__'):
            self.roles = getRoles(
                content, '__call__', content.__call__, self.roles)
        request.roles = self.roles

        return content





import raven.utils.wsgi
from .log import object_name, object_path


class RavenLoggingPlugin(object):

    def __init__(self, raven_config):
        self.client = raven.Client(**raven_config)

    def __call__(self, request, response, obj, exc_info, exc_text, extra):
        self.client.captureException(
            exc_info=exc_info,
            data={
                'message': "".join(exc_text),
                'sentry.interfaces.Http': {
                    'url': request.get('URL', 'n/a'),
                    'method': request.environ.get('REQUEST_METHOD', 'n/a'),
                    'query_string': request.environ.get('QUERY_STRING', 'n/a'),
                    'headers': dict(raven.utils.wsgi.get_headers(request.environ)),
                    'env': dict(raven.utils.wsgi.get_environ(request.environ))
                    }
                }, extra={
                'Object Class': object_name(obj),
                'Object Name': object_path(obj)
                })



# -*- coding: utf-8 -*-
# Copyright (c) 2011-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from zope.interface import Interface, Attribute, implements
from zope.publisher.interfaces.browser import IDefaultBrowserLayer

from ZPublisher.interfaces import IPubAfterTraversal, IPubEvent
from ZPublisher.pubevents import PubStart, PubSuccess, PubFailure, \
    PubBeforeCommit, PubAfterTraversal, PubBeforeAbort, \
    PubBeforeStreaming


class IRequest(IDefaultBrowserLayer):

    def query_plugin(context, iface):
        """Query an adapter that adapts context and request. Store it
        All request variable must set properlyfor future use.
        """

    def get_plugin(iface):
        """Return a previously queried plugin.
        """


class ITestRequest(IRequest):
    application = Attribute(u'Application root')
    method = Attribute(u"Queried method")
    path = Attribute(u"Queried path")


class IVirtualHosting(Interface):
    """Implement the virtual hosting.
    """
    request = Attribute(u"Request being published")
    context = Attribute(u"Application root")
    root = Attribute(u"Virtual Hosting that have been computed, if any.")

    def rewrite_url(base_url, original_url):
        """Rewrite ``original_url`` (a currently valid URL in the
        current computed virtual host) to be accessible via
        ``base_url`` (an another available virtual host). If
        ``base_url`` is empty, only a global path will be returned.
        """

    def __call__(method, path):
        """Return a tuple (root, method, path) to use after virtual
        hosting being done.

        All request variable must be updated properly.
        """


class ITraverser(Interface):
    """Traverse to the published content.
    """
    request = Attribute(u"Request being traversed")
    context = Attribute(u"Application root")

    def __call__(method, path):
        """Traversed to the published content.

        All request variable must be set properly.
        """


class IAuthenticator(Interface):
    """Do the Zope authentication, after the traverser have been called.
    """
    request = Attribute(u"Request")
    context = Attribute(u"Published content")

    def __call__(validation_hook=None):
        """Do the authentication. You can raise Unauthorized in case
        of failure.

        If validation_hook is provided, you must call it if a user is
        authenticated.

        All request variable must be set properly.
        """


# Publications Events, review and extend Zope2 publication events


class IPublicationAfterTraversal(IPubAfterTraversal):
    content = Attribute(u"Content traversed to")


class IPublicationAfterRender(IPubEvent):
    content = Attribute(u"Content rendered")


class IPublicationBeforeError(IPubEvent):
    content = Attribute(u"Content where the error happened.")


class PublicationStart(PubStart):
    pass


class PublicationAfterTraversal(PubAfterTraversal):
    implements(IPublicationAfterTraversal)

    def __init__(self, request, content):
        self.request = request
        self.content = content


class PublicationBeforeError(object):
    implements(IPublicationBeforeError)

    def __init__(self, request, content):
        self.request = request
        self.content = content

class PublicationAfterRender(object):
    implements(IPublicationAfterRender)

    def __init__(self, request, content):
        self.request = request
        self.content = content


class PublicationSuccess(PubSuccess):
    pass


class PublicationBeforeCommit(PubBeforeCommit):
    pass


class PublicationFailure(PubFailure):
    pass


class PublicationBeforeAbort(PubBeforeAbort):
    pass


class PublicationBeforeStreaming(PubBeforeStreaming):
    pass



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from setuptools import setup, find_packages
import os

version = '2.1'

tests_require = [
    'infrae.testing',
    'wsgi_intercept',
    'zope.site',
    'zope.testbrowser',
    ]

setup(name='infrae.wsgi',
      version=version,
      description="WSGI support for Zope 2",
      long_description=open("README.txt").read() + "\n" +
                       open(os.path.join("docs", "HISTORY.txt")).read(),
      classifiers=[
          "Environment :: Web Environment",
          "Intended Audience :: Developers",
          "License :: OSI Approved :: Zope Public License",
          "Programming Language :: Python",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Framework :: Zope2",
          ],
      keywords='zope2 wsgi silva infrae',
      author='Sylvain Viollon',
      author_email='info@infrae.com',
      url='http://svn.infrae.com/infrae.wsgi/trunk',
      license='ZPL',
      package_dir={'': 'src'},
      packages=find_packages('src'),
      namespace_packages=['infrae'],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
        'Zope2 >= 2.12.4',
        'five.grok',
        'setuptools',
        'zope.cachedescriptors',
        'zope.component',
        'zope.event',
        'zope.interface',
        'zope.processlifetime',
        'zope.publisher',
        'zope.security',
        ],
      entry_points={
        'paste.app_factory': [
            'zope2 = infrae.wsgi.paster:zope2_application_factory',
            ],
        'paste.filter_app_factory': [
            'threads = infrae.wsgi.debugging:make_middleware',
            ],
        },
      tests_require = tests_require,
      extras_require = {'test': tests_require},
      )



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt


def format_http_header(name):
    """This format an http header key.
    """
    return '-'.join(map(lambda s: s.capitalize(), name.split('-')))


class HTTPHeaders(dict):
    """Implement a case insensitive dictionary to store HTTP headers.
    """

    def __setitem__(self, key, value):
        dict.__setitem__(self, key.lower(), value)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.lower())

    def update(self, values):
        for key, value in values:
            dict.__setitem__(self, key.lower(), value)

    def get(self, key, default=None):
        return dict.get(self, key.lower(), default)

    def has_key(self, key):
        return dict.has_key(self, key.lower())

    __contains__ = has_key

    def items(self):
        items = []
        for key, value in dict.items(self):
            items.append((format_http_header(key), str(value)))
        return items

    def iteritems(self):
        for key, value in dict.iteritems(self):
            yield((format_http_header(key), str(value)))



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from infrae.wsgi.headers import HTTPHeaders

import unittest

class HeaderTestCase(unittest.TestCase):

    def test_set_and_get(self):
        """You can set and read values. This is case insensitive.
        """
        headers = HTTPHeaders()

        headers['content-type'] = 'text/plain'
        self.assertEquals(headers['content-type'], 'text/plain')
        self.assertEquals(headers['Content-Type'], 'text/plain')

        headers['Content-Type'] = 'application/octet-stream'
        self.assertEquals(headers['content-type'], 'application/octet-stream')
        self.assertEquals(headers['Content-Type'], 'application/octet-stream')

    def test_non_existing_get(self):
        """Looking a non existing key raises KeyError.
        """
        headers = HTTPHeaders()

        self.assertRaises(KeyError, headers.__getitem__, 'Missing')
        self.assertRaises(KeyError, headers.__getitem__, 'missing')

    def test_has_key(self):
        """You can check is an header is already set. This is case
        insensitive as well.
        """
        headers = HTTPHeaders()

        headers['Content-Length'] = '42'
        self.assertEquals(headers.has_key('Content-Length'), True)
        self.assertEquals(headers.has_key('content-length'), True)
        self.assertTrue('Content-Length' in headers)
        self.assertTrue('content-length' in headers)

        self.assertEquals(headers.has_key('Server'), False)
        self.assertEquals(headers.has_key('server'), False)
        self.assertFalse('Server' in headers)
        self.assertFalse('server' in headers)

    def test_items(self):
        """items return a list of tuples containing all headers. Header
        names are normalized, and value are converted to strings.
        """
        headers = HTTPHeaders()

        headers['content-type'] = 'text/plain'
        headers['Content-length'] = 42
        self.assertEquals(
            headers.items(),
            [('Content-Length', '42'), ('Content-Type', 'text/plain')])
        self.assertEquals(
            list(headers.iteritems()),
            [('Content-Length', '42'), ('Content-Type', 'text/plain')])


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(HeaderTestCase))
    return suite



# -*- coding: utf-8 -*-
# Copyright (c) 2010-2012 Infrae. All rights reserved.
# See also LICENSE.txt

from cgi import escape
from urllib import quote
import socket
import re

from ZPublisher.HTTPResponse import status_reasons
from ZPublisher.Iterators import IStreamIterator
from zope.publisher.interfaces.http import IResult
from zope.event import notify
import zExceptions

from infrae.wsgi.headers import HTTPHeaders
from infrae.wsgi.interfaces import PublicationBeforeStreaming
from infrae.wsgi.log import log_invalid_response_data

HEAD_REGEXP = re.compile('(<head[^>]*>)', re.I)
BASE_REGEXP = re.compile('(<base.*?>)',re.I)


class AbortPublication(Exception):
    """Exception to abort all the publication process.
    """

    def __init__(self, started=True):
        self.response_started = started


class StreamIteratorIterator(object):
    """Make a IStreamIterator a real iterator, because it lack of an
    __iter__ method ...
    """

    def __init__(self, stream):
        assert IStreamIterator.providedBy(stream)
        self.__stream = stream

    def __iter__(self):
        return self.__stream


def format_cookies(cookies):
    """Format cookies as WSGI HTTP headers.
    """
    formatted_cookies = []
    for name, options in cookies.items():
        cookie = '%s="%s"' % (name, quote(options['value']))
        for key, value in options.items():
            key = key.lower()
            if key == 'expires':
                cookie = '%s; Expires=%s' % (cookie, value)
            elif key == 'domain':
                cookie = '%s; Domain=%s' % (cookie, value)
            elif key == 'path':
                cookie = '%s; Path=%s' % (cookie, value)
            elif key == 'max_age':
                cookie = '%s; Max-Age=%s' % (cookie, value)
            elif key == 'comment':
                cookie = '%s; Comment=%s' % (cookie, value)
            elif key == 'secure' and value:
                cookie = '%s; Secure' % cookie
            # Some browsers recognize this cookie attribute
            # and block read/write access via JavaScript
            elif key == 'http_only' and value:
                cookie = '%s; HTTPOnly' % cookie
        formatted_cookies.append(('Set-Cookie', cookie))
    return formatted_cookies


class WSGIResponse(object):
    """A response object using a WSGI connection

    This Response object knows nothing about ZServer, but tries to be
    compatible with the ZPublisher.HTTPResponse.
    """
    default_charset = 'utf-8'
    realm = 'Zope'

    # This is just need by FSDTMLFile. It should be removed when no
    # DTML files are needed anymore
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, environ, start_response, debug_mode=False):
        self.headers = HTTPHeaders()
        self.status = 200
        self.cookies = {}
        self.body = None
        self.debug_mode = debug_mode
        self.__base = None
        self.__environ = environ
        self.__start_response = start_response
        self.__started = False
        self.__write = None

    def redirect(self, location, status=302):
        self.status = status
        self.headers['Location'] = location

    def write(self, data):
        # This is deprecated, please return an iterable instead of
        # using write()
        if self.__write is None:
            self.__write = self.startWSGIResponse(stream=True)
        try:
            if isinstance(data, unicode):
                data = data.encode(self.default_charset)
            if not isinstance(data, str):
                log_invalid_response_data(data, self.__environ)
            self.__write(data)
        except (socket.error, IOError):
            # If we can't write anymore to the socket, abort all the
            # publication process.
            raise AbortPublication(started=True)

    def setBase(self, base):
        # HTTPResponse compatibility: need to randomly insert a base
        # header in HTML.
        if base and not base.endswith('/'):
            base += '/'
        self.__base = base

    def __insertBase(self, body):
        # HTTPResponse compatibility: insert a base tag if
        # needed. There is no way to get ride of that, several ZMI
        # screens rely on this (form using :method variables).
        content_type = self.headers.get('content-type', '').split(';')[0]
        if content_type != 'text/html':
            return body

        if self.__base:
            if body:
                match = HEAD_REGEXP.search(body)
                if match is not None:
                    index = match.start(0) + len(match.group(0))
                    ibase = BASE_REGEXP.search(body)
                    if ibase is None:
                        return '%s\n<base href="%s" />\n%s' % (
                            body[:index],
                            escape(self.__base, 1),
                            body[index:])
        return body

    def setBody(self, body, **options):
        # We ignore options, but do __insertBase if body is a string.
        if isinstance(body, basestring):
            body = self.__insertBase(body)
            if isinstance(body, unicode):
                body = body.encode(self.default_charset)
        self.body = body

    def getBody(self):
        return self.body

    def setStatus(self, status, msg=None):
        # We ignore msg and use our own.
        self.status = status

    def getStatus(self):
        return self.status

    def setHeader(self, name, value, literal=0):
        # literal=0 is for HTTPResponse compatibility
        self.headers[name] = value

    addHeader = setHeader

    def getHeader(self, name, literal=0):
        # literal=0 is for HTTPResponse compatibility
        return self.headers.get(name)

    def appendCookie(self, name, value):
        name = str(name)
        value = str(value)

        cookies = self.cookies
        if cookies.has_key(name):
            cookie = cookies[name]
        else:
            cookie = cookies[name] = {}
        if cookie.has_key('value'):
            cookie['value'] = '%s:%s' % (cookie['value'], value)
        else:
            cookie['value'] = value

    def expireCookie(self, name, **options):
        options['max_age'] = 0
        options['expires'] = 'Wed, 31-Dec-97 23:59:59 GMT'
        self.setCookie(name, 'deleted', **options)

    def setCookie(self, name, value, **options):
        name = str(name)
        value = str(value)

        cookies = self.cookies
        if cookies.has_key(name):
            cookie = cookies[name]
        else:
            cookie = cookies[name] = {}
        for cookie_key, cookie_value in options.items():
            cookie[cookie_key] = cookie_value
        cookie['value'] = value

    def startWSGIResponse(self, stream=False):
        if self.__started:
            return self.__write
        self.__started = True

        # If the body is an IResult, it is a case of streaming where
        # we don't fix headers.
        stream = stream or IResult.providedBy(self.body)

        if not stream:
            # If we are not streaming, we try to set Content-Length,
            # Content-Type and adapt status if there is no content.
            if not self.headers.has_key('Content-Length'):
                content_length = None
                if (isinstance(self.body, basestring) or
                    IStreamIterator.providedBy(self.body)):
                    content_length = len(self.body)
                elif self.body is None:
                    content_length = 0
                if content_length is not None:
                    self.headers['Content-Length'] = content_length

            if self.headers.has_key('Content-Length'):
                content_length = self.headers['Content-Length']
                if not content_length and self.status == 200:
                    # Set no content status if there is no content
                    self.status = 204

            if not self.headers.has_key('Content-Type'):
                if content_length:
                    # If there is content and no Content-Type, set HTML
                    self.headers['Content-Type'] = 'text/html;charset=%s' % (
                        self.default_charset,)
        else:
            # Fire event before streaming
            notify(PublicationBeforeStreaming(self))

            # Fix default Content-Type
            if not self.headers.has_key('Content-Type'):
                self.headers['Content-Type'] = 'text/html;charset=%s' % (
                    self.default_charset,)


        formatted_status = "%d %s" % (
            self.status, status_reasons.get(self.status, 'OK'))
        formatted_headers = self.headers.items() + format_cookies(self.cookies)
        return self.__start_response(formatted_status, formatted_headers)

    def getWSGIResponse(self):
        # This return a tuple (data_sent, data_to_send_to_WSGI)
        result = self.body
        if result is not None:
            # If we have an iterator, we say that data have been sent
            # in order not to commit the transaction finish consuming
            # the iterator.
            if IStreamIterator.providedBy(result):
                return (True, StreamIteratorIterator(result))
            elif IResult.providedBy(result):
                return (True, result)
            if not isinstance(result, str):
                log_invalid_response_data(result, self.__environ)
            return (self.__started, [result,])
        return (self.__started, [])

    def _unauthorized(self):
        # Unauthorized is implemented like this to be compliant with
        self.setHeader(
            'WWW-Authenticate', 'basic realm="%s"' % self.realm)

    # ZPublisher define a method for each error on the response
    # object. We need to add them to be compatible

    def unauthorized(self):
        raise zExceptions.Unauthorized("Please authenticate")

    def debugError(self, name):
        import pdb; pdb.set_trace()
        raise zExceptions.NotFound(name)

    def forbiddenError(self, name):
        raise zExceptions.Forbidden(name)

    def badRequestError(self, name):
        raise zExceptions.BadRequest(name)

    def notFoundError(self, name):
        raise zExceptions.NotFound(name)

    # Support for ConflictError/Retry. We reuse the same response.

    def retry(self):
        return self
